/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef REALPHOTONICSDRIVER_H
#define REALPHOTONICSDRIVER_H
#include <string>
#include <vector>
#define _GNULINUX

#include "Utilities/Counter.h"

#include </usr/local/src/baumer/inc/bgapi2_genicam/bgapi2_genicam.hpp>
#include "drivers/gigecam/photonics/BufferInformation.h"
#include "drivers/gigecam/photonics/CameraHandler.h"
#include "drivers/gigecam/photonics/DoubleBufferHandler.h"
#include "drivers/gigecam/photonics/PhotonicsDef.h"
#include "drivers/gigecam/photonics/PhotonicsState.h"

namespace photonics {

/*!
 * \class RealPhotonicsDriver
 * \brief Real implementation class for the Psl device driver
 *
 * This class is a real implementation of Psl device driver.
 */
class RealPhotonicsDriver: public PhotonicsState{

public:

	/*!
	 * \brief Constructor
	 * \param[in] owner The device driver main class link
	 */
	RealPhotonicsDriver(PhotonicsDriver* owner);

	/*!
	 * \brief Destructor
	 */
	virtual ~RealPhotonicsDriver();

	/*!
	 * \brief Init command implementation
	 */
	virtual void init();

	/*!
	 * \brief Clear command implementation
	 */
	virtual void clear();

	/*!
	 * \brief Write Param command implementation
	 */
	virtual void writeParam();

	/*!
	 * \brief Read command implementation
	 */
	virtual void read();

	/*!
	 * \brief Synchronize Read command implementation
	 */
	virtual void synchroniseRead();

	/*!
	 * \brief Start command implementation
	 */
	virtual void start();

	/*!
	 * \brief Resume command implementation
	 */
	virtual void resume();

	/*!
	 * \brief Pause command implementation
	 */
	virtual void pause();

	/*!
	 * \brief Stop command implementation
	 */
	virtual void stop();

	/*!
	 * \brief Read Infos command implementation
	 */
	virtual void readStatus();

	/*!
	 * \brief Read Infos command implementation
	 */
	virtual void readInfos();

	/*!
	 * \brief Regroup command implementation
	 */
	//	virtual void regroup();
//	int32 GetAvailablePhotonicseras();
//	std::string getConfigFile(BGAPI2::Device* device);

private:




	// a flag which controls the status command
	std::atomic<bool> command_status;
	// a flag which controls the calculation command
	std::atomic<bool> command_calculation;

	// this function activates the feature command on all connected cameras
	// the feature command use the ExposureTime feature of the camera
	// the parameter passes the new exposure time to be written to the cameras
	void StartFeatureCommandExposure(CCameraHandler * camera_handler, double exposure_time);

	// this function activates the capture command on all connected cameras
	// the parameter passes the number of images to be captured
	void StartCaptureCommand(CCameraHandler * camera_handler, unsigned int images);

	// -------------
	// thread control, finish flag and thread functions
	// -------------

	// a flag which is used to stop all the used threads at once
	bool finish_console_threads = false;

	// the thread routine of the status command is used to print status information about the current buffer,
	// exposure time, logging information and calculation results
	// prints the status information to the console
	void StatusThreadRoutine(CCameraHandler* camera_handler);

	// the thread routine of the calculation command is used to make a image calculation
	// prints the calculation result to the console
	void CalculationThreadRoutine(CCameraHandler* camera_handler);

	// -------------
	// declaration of synchronisation functions needed for synchronized buffer access between the application threads
	// -------------

	// this function is used to check if new data is available
	bool HasDoubleBufferHandlerNewData(CDoubleBufferHandler* buffer_handler);

	// this function is used within the thread functions to receive a new buffer
	BGAPI2::Buffer* PullBufferFromDoubleBufferHandler(CDoubleBufferHandler* buffer_handler);

	// this function is used within the thread functions to return a buffer
	void FreeBufferToDoubleBufferHandler(CDoubleBufferHandler* buffer_handler, BGAPI2::Buffer* buffer);

	// structure with the buffer object and a reference counter
	// the reference counter is incremented with every call to PullBufferFromBufferModule
	// and decremented with every call to FreeBufferToBufferModule
	// if the reference counter reaches 0 the buffer will return to BGAPI
	// the new data flag is available for all application threads
	// this flags makes sure that one application thread of one camera process a camera buffer only once
	struct CameraBuffer{
	    CameraBuffer() {
	        buffer_ = nullptr;
	        ref_counter_ = 0;
	    }
	    std::mutex lock_;
	    BGAPI2::Buffer* buffer_;
	    std::atomic<int> ref_counter_;
	    std::map<std::thread::id, std::atomic<bool>> new_data;
	};
	// this map manages all reference counter and buffer objects for each camera
	std::map<CDoubleBufferHandler*, CameraBuffer> current_buffer;

	// result of calculation thread stored in a simple map of string
	std::map<CDoubleBufferHandler*, std::string> calculation_result;

	// mutex to synchronize the access to the calculation result
	std::mutex calculation_result_lock;

	// ---------------
	// helper function, used to reduce the length of some functions
	// ---------------

	// initialize buffer structure, which is used for buffer and data exchange between
	// the main and application threads
	void InitCurrentBufferStruct(CCameraHandler *camera_handler, std::list<std::thread::id> thread_id_list);

	// waits until all connected cameras are streaming
	// prints the streaming status, camera name and serial number to console
	void WaitUntilStreamingIsActive(CCameraHandler * camera_handler);

	// function used in automatic mode to wait for all camera buffer
	int WaitOfImages(CCameraHandler * camera_handler, unsigned int number_of_images);

	// ---------------
	// example application functions used by the application threads
	// ---------------

	// perform a camera depended the logging output used by the status thread
	void PrintLoggingInformation(CCameraControl * camera_control);

	// perform the status output used by the status thread
	void PrintCameraStatus(CCameraControl * camera_control);

	// print calculation results used by status thread
	void PrintCalculationStatus(CCameraControl * camera_control);

	// perform a mean value calculation used by the calculation thread
	void DoCalculation(CCameraControl * camera_control);



		// instantiate the camera handler
    CCameraHandler m_camera_handler;

    // instantiate the buffer handler
    CDoubleBufferHandler m_double_buffer;

    // structure with the buffer object and a reference counter
    // the reference counter is incremented with every call to PullBufferFromBufferModule
    // and decremented with every call to FreeBufferToBufferModule
    // if the reference counter reaches 0 the buffer will return to BGAPI
    // the new data flag is available for all application threads
     // this map manages all reference counter and buffer objects for each camera
    std::map<CDoubleBufferHandler*, CameraBuffer> m_current_buffer;

    bool m_finish_console_threads;

    // a flag which controls the status command
    std::atomic<bool> m_command_status;

    int32 m_return_code;

	Counter m_TimeCounter;

};

}
#endif //REALPSLDRIVER_H
